from .env import DATA_FOLDER
import pandas as pd
from typing import Dict
import os
from os.path import exists
from huggingface_hub import HfApi
from numpy import ndarray
import numpy as np
from .datafolder import DataFolder

def import_dataset(name:str, data:str, force_download:bool=False)->ndarray:

    """Finds the dataset with the given name for the given type of data.

    Parameters
    ----------

    name : str
        Name of the dataset to find.
    data : str
        Name of the type of data to find the dataset for (structure or DMS).
    force_download : bool
        Whether to force download the dataset from HuggingFace Hub. Defaults to False.

    Returns
    -------

    ndarray
        The dataset with the given name for the given type of data.

    Example
    -------

    >>> import_dataset(name='for_testing', data='structure').shape
    (2,)
    >>> import_dataset(name='for_testing', data='DMS').shape
    (2,)
    >>> import_dataset(name='for_testing', data='structure', force_download=True).shape
    (2,)
    >>> import_dataset(name='for_testing', data='DMS', force_download=True).shape
    (2,)

    """

    assert data in ['structure', 'DMS'], "data must be either 'structure' or 'DMS'"

    # Get the data folder
    try:
        assert not force_download, "Force download from HuggingFace Hub"
        datafolder = DataFolder.from_local(name=name)
    except:
        try:
            datafolder = DataFolder.from_huggingface(name=name)
        except:
            raise ValueError("Dataset not found on HuggingFace Hub")

    # Get the dataset
    if data == 'structure':
        if not exists(datafolder.get_base_pairs_npy()) or force_download:
            datafolder.generate_npy()
        return np.load(datafolder.get_base_pairs_npy(), allow_pickle=True)

    elif data == 'DMS':
        if not exists(datafolder.get_dms_npy()) or force_download:
            datafolder.generate_npy()
        return np.load(datafolder.get_dms_npy(), allow_pickle=True)




from typing import Any, List
from .parsers import *
from .util import add_braces_if_no_braces, dot2int, int2dot, seq2int
import numpy as np

class Datapoint:

    def __init__(self, sequence, reference, structure=None, dms=None, paired_bases=None):
        for attr in [sequence, reference]:
            assert isinstance(attr, str), f"Expected {attr} to be a string, got {type(attr)} instead."
            assert len(attr) > 0, f"Expected {attr} to be non-empty."
        self.reference = reference
        self.sequence = sequence
        self.structure = structure
        self.paired_bases = paired_bases if paired_bases is not None else self.structure_to_paired_bases(structure) if structure is not None else None
        self.dms = dms
        self.opt_dict = {'paired_bases': self.paired_bases, 'dms': self.dms}

    def to_dict(self):
        return self.reference, {'sequence': self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}

    def to_flat_dict(self):
        return {'reference': self.reference, 'sequence': self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}

    def __str__(self):
        return '"'+self.reference+'"' + ':' + str({"sequence": self.sequence, **{k:v for k,v in self.opt_dict.items() if v is not None}}).replace("'",'"').replace('(','[').replace(')',']').replace('None','null')

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.reference}', sequence='{self.sequence}', paired_bases={self.paired_bases}, dms={self.dms})"

    def structure_to_paired_bases(self, structure):
        """Returns a list of tuples (i,j) where i and j are paired bases."""
        paired_bases = []
        stack = []
        for i, char in enumerate(structure):
            if char == '(':
                stack.append(i)
            elif char == ')':
                paired_bases.append((stack.pop(), i))
        return paired_bases

    def embed_sequence(self):
        """Returns a list of integers corresponding to the sequence.

        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint.embed_sequence()
        array([1, 1, 2, 2, 3, 3])
        """
        return np.array([seq2int[base] for base in self.sequence])

    def embed_structure(self):
        """Returns a list of integers corresponding to the structure.

        >>> datapoint = Datapoint(reference='reference', sequence='AACCGG', structure='((..))', dms=[1.0, 2.0, 3.0])
        >>> datapoint.embed_structure()
        array([2, 2, 1, 1, 3, 3])
        """
        return np.array([dot2int[base] for base in self.structure])



class DatapointFactory:
    """A datapoint is a tuple (sequence, reference, structure, dms) where:

    - reference is the name of the sequence in the fasta file
    - sequence is a string of A, C, G, U
    - structure is a string of (, ), .
    - dms is a list of floats corresponding to the reactivity of each base in the sequence

    """

    def __call__(sequence, reference, structure=None, dms=None, predict_structure=False, predict_dms=False):
        if structure is None and predict_structure:
            structure = Fasta.predict_structure(sequence)
        return Datapoint(sequence, reference, structure, dms)

    def from_ct(ct_file, predict_dms):
        """Create a datapoint from a ct file. If predict_dms is True, the dms will be predicted using RNAstructure"""
        reference, sequence, paired_bases = Ct.parse(ct_file)
        return Datapoint(sequence, reference, paired_bases=paired_bases, dms = Ct.predict_dms(ct_file) if predict_dms else None)

    def from_fasta(sequence, reference, predict_structure, predict_dms):
        """Create a datapoint from a fasta file. The structure and dms will be None."""
        structure, dms = None, None
        if predict_structure:
            structure = Fasta.predict_structure(sequence)
        if predict_dms:
            dms = Fasta.predict_dms(sequence)
        return Datapoint(sequence, reference, structure, dms)

    def from_dreem_output(reference, sequence, mutation_rate, predict_structure):
        """Create a datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        return Datapoint(
            sequence=sequence,
            reference=reference,
            structure=Fasta.predict_structure(sequence) if predict_structure else None,
            dms=mutation_rate)


    def from_json_line(string):
        """Create a datapoint from a json line. The json line should have the following format:
        "reference": {"sequence": "sequence", "paired_bases": [[1, 2], [3,4]], "dms": [1.0, 2.0, 3.0]}

        Example:
        >>> DatapointFactory.from_json_line('"reference": {"sequence": "sequence"}')
        Datapoint('reference', sequence='sequence', paired_bases=None, dms=None)
        >>> DatapointFactory.from_json_line('"reference": {"sequence": "sequence", "paired_bases": [[1, 2], [3, 4]], "dms": [1.0, 2.0, 3.0]}')
        Datapoint('reference', sequence='sequence', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])
        """
        string= string.strip()[:-1] if string.strip()[-1] == ',' else string.strip()
        string = add_braces_if_no_braces(string)
        d = json.loads(string)
        ref = list(d.keys())[0]
        d = d[ref]
        return Datapoint(
            reference=ref,
            sequence=d['sequence'],
            paired_bases=d['paired_bases'] if 'paired_bases' in d else None,
            dms=d['dms'] if 'dms' in d else None
        )






import os
from .util import source_env

path_env = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'env')

try:
    vars_env = source_env(path_env)
except FileNotFoundError:
    print("No env file found. Using environment variables.")
    vars_env = {}

HUGGINGFACE_TOKEN = os.environ.get('HUGGINGFACE_TOKEN')
DATA_FOLDER = os.environ.get('DATA_FOLDER')
DATA_FOLDER_TESTING = os.environ.get('DATA_FOLDER_TESTING')
RNASTRUCTURE_PATH = os.environ.get('RNASTRUCTURE_PATH')
RNASTRUCTURE_TEMP_FOLDER = os.environ.get('RNASTRUCTURE_TEMP_FOLDER')

for var in ["HUGGINGFACE_TOKEN", "DATA_FOLDER", "DATA_FOLDER_TESTING", "RNASTRUCTURE_PATH", "RNASTRUCTURE_TEMP_FOLDER"]:
    if exec(var) == None and var in vars_env and vars_env[var] != None:
        exec(f"{var} = vars_env['{var}']") # read from env file directly
    else:
        raise ValueError(f"Environment variable {var} is not set. Please run `source env`.")





from typing import Any

import json
from .env import DATA_FOLDER

from .path import PathDatafolder
from .env import DATA_FOLDER, HUGGINGFACE_TOKEN
from .list_datapoints import ListofDatapoints
from .info_file import infoFileWriter
import os
import numpy as np
from huggingface_hub import HfApi
from huggingface_hub import snapshot_download

GENERATE_NPY = False
PREDICT_STRUCTURE = False
PREDICT_DMS = False
ROUSKINLAB = 'rouskinlab/'

class DataFolderTemplate(PathDatafolder):

    def __init__(self, name, path_out) -> None:
        super().__init__(name, path_out)
        self.name = name
        self.datapoints = ListofDatapoints([])

    def generate_npy(self):
        """Generate the npy files:
            - reference.npy
            - sequence.npy
            - base_pairs.npy
            - dms.npy

        Each file corresponds to a key in the datapoints dictionary contained in the json file.
        For example, if the 1st line of the json file is:
        {"RF02271.fa.csv_1":{"sequence": "AAACCCCAGUAGGGGUUU", "paired_bases": [[6, 11], [5, 12], [4, 13], [3, 14], [2, 15], [1, 16], [0, 17]], "dms": [0.3317625732102161, 0.8619477949493498, 0.9849505502538177, 0.9997828311396947, 0.9999562120800795, 0.9999454042235505, 0.9970246205212348, 3.298876877182004e-10, 1.0113344525791075e-08, 9.30187856307975e-09, 2.1481183689942838e-10, 0.996953491189411, 0.9999737756512344, 0.9999846271614127, 0.9997971647502685, 0.9848757245213886, 0.8618760196213241, 0.3319091664146036]}}
        then the 1st line of the reference.npy file will be:
        AAACCCCAGUAGGGGUUU
        """

        self.datapoints.to_reference_npy(self.get_references_npy())
        self.datapoints.to_sequence_npy(self.get_sequences_npy())

        d1 = self.datapoints.datapoints[0]

        if hasattr(d1, 'paired_bases'):
            self.datapoints.to_base_pairs_npy(self.get_base_pairs_npy())

        if hasattr(d1, 'dms'):
            self.datapoints.to_dms_npy(self.get_dms_npy())

class CreateDatafolderTemplate(DataFolderTemplate):

    def __init__(self, path_in, path_out, name, source, predict_structure, predict_dms) -> None:
        name = self._set_name(name, path_in)
        super().__init__(name, path_out)
        self.name = name
        self.path_in = path_in
        self.path_out = path_out
        self.api = HfApi(token=HUGGINGFACE_TOKEN)
        self.predict_structure = predict_structure
        self.predict_dms = predict_dms

        # move path_in to source folder
        os.makedirs(self.get_source_folder(), exist_ok=True)
        os.system(f'cp -fr {path_in} {self.get_source_folder()}')

        # Write info file
        infoFileWriter(source=source, datafolder=self).write()


    def __repr__(self) -> str:
        return f"{self.__class__.__name__} @{self.get_main_folder()}"

    def _set_name(self, name, path_in):
        if name is None:
            name = path_in.replace('\\','/').split('/')[-1].split('.')[0]
        return name

    def create_repo(self, exist_ok=False, private=True):

        """Create a repo on huggingface.co.

        Parameters
        ----------

        exist_ok : bool, optional
            If True, the repo is created even if it already exists. Default is False.

        private : bool, optional
            If True, the repo is private. Default is True.

        Examples
        --------

        >>> datafolder = DataFolder.from_dreem_output(path_in='data/input_files_for_testing/dreem_output.json')
        >>> datafolder.create_repo(exist_ok=True)

        """

        self.api.create_repo(
            repo_id=ROUSKINLAB+self.name,
            token=HUGGINGFACE_TOKEN,
            exist_ok=exist_ok,
            private=private,
            repo_type="dataset",
        )

    def upload_folder(self, revision = 'main', commit_message=None, commit_description=None, multi_commits=False, run_as_future=False, **kwargs):

        """Upload the datafolder to huggingface.co. Only the json file, the README.md and the source files are uploaded.

        Parameters
        ----------

        revision : str, optional
            The revision of the repo (i.e. the branch on the HuggingFace hub). Default is 'main'.

        commit_message : str, optional
            The commit message. Default is None.

        commit_description : str, optional
            The commit description. Default is None.

        multi_commits : bool, optional
            Useful for large files. If True, the upload is done in multiple commits. Default is False.

        run_as_future : bool, optional
            If True, the upload is run asynchonously. The state of the upload can be checked with the returned future. See sample code. Default is False.

        Examples
        --------

        >>> datafolder = DataFolder.from_dreem_output(path_in='data/input_files_for_testing/dreem_output.json')
        >>> datafolder.upload_folder('main', 'pytest commit', 'A commit generated by pytest for the upload_folder method')
        >>> future = datafolder.upload_folder(run_as_future=True)
        >>> future.result() # wait for the upload to finish
        'https://huggingface.co/datasets/rouskinlab/dreem_output/tree/main/'
        >>> future.done() # check if the upload is done
        True

        """

        future = self.api.upload_folder(
            repo_id=ROUSKINLAB+self.name,
            folder_path=self.get_main_folder(),
            repo_type="dataset",
            token=HUGGINGFACE_TOKEN,
            revision=revision,
            commit_message=commit_message,
            commit_description=commit_description,
            multi_commits=multi_commits,
            run_as_future=run_as_future,
            allow_patterns=["source/*", "info.json", "data.json", "README.md"],
            **kwargs
        )

        if run_as_future:
            return future





    def dump_datapoints(self, generate_npy):
        """Dump the datapoints to a json file."""

        if not os.path.exists(self.get_json()):
            self.datapoints.to_json(self.get_json())

        if generate_npy:
            self.generate_npy()



class CreateDatafolderFromDreemOutput(CreateDatafolderTemplate):

    """Create a datafolder from a dreem output file.

    Parameters
    ----------

    path_in : str
        path_in to the dreem output file.

    name : str, optional
        Name of the datafolder. If None, the name is the name of the file or folder.

    predict_structure : bool, optional
        If True, the structure of the RNA is predicted. Default is True.

    generate_npy : bool, optional
        If True, the npy files are generated. Default is True.


    Examples
    --------

    >>> datafolder = DataFolder.from_dreem_output(path_in='data/input_files_for_testing/dreem_output.json', generate_npy=True)
    >>> datafolder.name
    'dreem_output'
    >>> print(str(datafolder).split(' ')[0])
    CreateDatafolderFromDreemOutput
    >>> os.path.isfile(datafolder.get_json())
    True
    """

    def __init__(self, path_in, path_out, name, predict_structure, generate_npy) -> None:
        super().__init__(path_in, path_out, name, source = 'dreem_output', predict_structure = predict_structure, predict_dms = False)

        self.datapoints = ListofDatapoints.from_dreem_output(path_in, predict_structure = predict_structure)
        self.dump_datapoints(generate_npy)


class CreateDatafolderFromFasta(CreateDatafolderTemplate):

    """ Create a datafolder from a fasta file.

    Parameters
    ----------

    path_in : str
        path_in to the dreem output file.

    name : str, optional
        Name of the datafolder. If None, the name is the name of the file or folder.

    predict_structure : bool, optional
        If True, the structure of the RNA is predicted. Default is True.

    predict_dms : bool, optional
        If True, the dms of the RNA is predicted. Default is True.

    generate_npy : bool, optional
        If True, the npy files are generated. Default is True.


    Examples
    --------

    >>> datafolder = DataFolder.from_fasta(path_in='data/input_files_for_testing/sequences.fasta', generate_npy=True)
    >>> datafolder.name
    'sequences'
    >>> print(str(datafolder).split(' ')[0])
    CreateDatafolderFromFasta
    >>> os.path.isfile(datafolder.get_json())
    True
    """

    def __init__(self, path_in, path_out, name, predict_structure, predict_dms, generate_npy) -> None:
        super().__init__(path_in, path_out, name, source = 'fasta', predict_structure = predict_structure, predict_dms = predict_dms)

        self.datapoints = ListofDatapoints.from_fasta(path_in, predict_structure = predict_structure, predict_dms = predict_dms)
        self.dump_datapoints(generate_npy)



class CreateDatafolderFromCTfolder(CreateDatafolderTemplate):

    """ Create a datafolder from a folder of ct files.

    Parameters
    ----------

    path_in : str
        path_in to the dreem output file.

    name : str, optional
        Name of the datafolder. If None, the name is the name of the file or folder.

    predict_dms : bool, optional
        If True, the dms of the RNA is predicted. Default is True.

    generate_npy : bool, optional
        If True, the npy files are generated. Default is True.


    Examples
    --------

    >>> datafolder = DataFolder.from_ct_folder(path_in='data/input_files_for_testing/ct_files', generate_npy=True)
    >>> datafolder.name
    'ct_files'
    >>> print(str(datafolder).split(' ')[0])
    CreateDatafolderFromCTfolder
    >>> os.path.isfile(datafolder.get_json())
    True
    """

    def __init__(self, path_in, path_out, name, predict_dms, generate_npy) -> None:
        super().__init__(path_in, path_out, name, source = 'ct', predict_structure = False, predict_dms = predict_dms)

        ct_files = [os.path.join(path_in, f) for f in os.listdir(path_in) if f.endswith('.ct') or f.endswith('.txt')]
        self.datapoints = ListofDatapoints.from_ct(ct_files, predict_dms = predict_dms)
        self.datapoints.filter_duplicates()
        self.dump_datapoints(generate_npy)


class LoadDatafolder(DataFolderTemplate):

    def __init__(self, name, root) -> None:
        super().__init__(name, root)


class LoadDatafolderFromHF(LoadDatafolder):
    """Load a datafolder from HuggingFace.

    Parameters
    ----------

    name : str
        Name of the datafolder.

    path_out : str
        Path to the folder where the datafolder is saved.

    revision : str, optional
        Revision of the datafolder. Default is 'main'.


    Examples
    --------

    >>> datafolder = LoadDatafolderFromHF(name='for_testing', path_out='data/datafolders')
    >>> datafolder.name
    'for_testing'
    """

    def __init__(self, name, path_out, revision='main') -> None:
        super().__init__(name, path_out)

        # Download the datafolder #TODO : check if the datafolder is already downloaded
        snapshot_download(
            repo_id = ROUSKINLAB+self.name,
            repo_type='dataset',
            local_dir=self.get_main_folder(),
            revision=revision,
            token=HUGGINGFACE_TOKEN,
            allow_patterns=['info.json', 'data.json']
            )

        assert os.path.isdir(self.get_main_folder()), f'No folder found in {self.get_main_folder()}'
        assert os.path.isfile(self.get_json()), f'No json file found in {self.get_main_folder()}'

        self.datapoints = ListofDatapoints.from_json(self.get_json())


class LoadDatafolderFromLocal(LoadDatafolder):
    """Load a datafolder from local.

    Parameters
    ----------

    name : str
        Name of the datafolder.

    path : str
        Path to the folder where the datafolder is saved. Defaut is set in the env variable DATAFOLDER_PATH.

    Examples
    --------

    >>> datafolder = LoadDatafolderFromLocal(name='for_testing', path='data/datafolders')
    >>> datafolder.name
    'for_testing'
    """

    def __init__(self, name, path) -> None:
        super().__init__(name, path)

        assert os.path.isdir(self.get_main_folder()), f'No folder found in {self.get_main_folder()}'
        assert os.path.isfile(self.get_json()), f'No json file found in {self.get_main_folder()}'

        self.datapoints = ListofDatapoints.from_json(self.get_json())

class DataFolder:
    """Create a datafolder from a fasta file, a json file or a folder of ct files.

    Parameters
    ----------

    path_in : str
        path_in to the fasta file, the json file or the folder of ct files.

    path_out : str
        path_out to the folder where the datafolder is created.

    name : str, optional
        Name of the datafolder. If None, the name is the name of the file or folder.

    predict_structure : bool, optional
        If True, the structure is predicted. Default is True.

    predict_dms : bool, optional
        If True, the dms is predicted. Default is True.


    Example
    -------

    >>> datafolder = DataFolder.from_fasta('data/input_files_for_testing/sequences.fasta')
    >>> datafolder = DataFolder.from_dreem_output('data/input_files_for_testing/dreem_output.json')
    >>> datafolder = DataFolder.from_ct_folder('data/input_files_for_testing/ct_files')
    >>> datafolder = DataFolder.from_huggingface('for_testing')
    >>> datafolder = DataFolder.from_local('for_testing')
    """

    def from_fasta(path_in, path_out=DATA_FOLDER, name = None, predict_structure = PREDICT_STRUCTURE, predict_dms = PREDICT_DMS, generate_npy = GENERATE_NPY)->CreateDatafolderFromFasta:
        """Create a datafolder from a fasta file. See CreateDatafolderFromFasta for more details."""
        return CreateDatafolderFromFasta(path_in, path_out, name, predict_structure, predict_dms, generate_npy)

    def from_dreem_output(path_in, path_out=DATA_FOLDER, name = None, predict_structure = PREDICT_STRUCTURE, generate_npy = GENERATE_NPY)->CreateDatafolderFromDreemOutput:
        """Create a datafolder from a dreem output file. See CreateDatafolderFromDreemOutput for more details."""
        return CreateDatafolderFromDreemOutput(path_in, path_out, name, predict_structure, generate_npy)

    def from_local(name, path=DATA_FOLDER)->LoadDatafolderFromLocal:
        """Load a datafolder from local. See LoadDatafolderFromLocal for more details."""
        return LoadDatafolderFromLocal(name, path)

    def from_ct_folder(path_in, path_out=DATA_FOLDER, name = None, predict_dms = PREDICT_DMS, generate_npy = GENERATE_NPY)->CreateDatafolderFromCTfolder:
        """Create a datafolder from a folder of ct files. See CreateDatafolderFromCTfolder for more details."""
        return CreateDatafolderFromCTfolder(path_in, path_out, name, predict_dms, generate_npy)

    def from_huggingface(name, path_out=DATA_FOLDER)->LoadDatafolderFromHF:
        """Load a datafolder from HuggingFace. See LoadDatafolderFromHF for more details."""
        return LoadDatafolderFromHF(name, path_out)




import os
import numpy as np
import json
import re
from .datapoint import Datapoint, DatapointFactory
from typing import List, Tuple, Union, Optional
from .parsers import Fasta, DreemOutput
import pandas as pd

class ListofDatapoints:

    def __init__(self, datapoints):
        self.datapoints = datapoints

    def __call__(self) -> List[Datapoint]:
        return self.datapoints

    @classmethod
    def from_fasta(cls, fasta_file, predict_structure, predict_dms):
        """Create a list of datapoint from a fasta file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        sequences, references = Fasta.parse(fasta_file)
        return cls([DatapointFactory.from_fasta(sequence, reference, predict_structure, predict_dms)
                    for sequence, reference in zip(sequences, references)])

    @classmethod
    def from_ct(cls, ct_files, predict_dms):
        """Create a list of datapoint from a ct file. The dms will be predicted if predict_dms is True."""
        return cls([DatapointFactory.from_ct(ct_file, predict_dms) for ct_file in ct_files])

    @classmethod
    def from_dreem_output(cls, dreem_output_file, predict_structure):
        """Create a list of datapoint from a dreem output file. The structure and dms will be predicted if predict_structure and predict_dms are True."""
        return cls([DatapointFactory.from_dreem_output(reference, sequence, mutation_rate, predict_structure)
            for reference, sequence, mutation_rate in DreemOutput.parse(dreem_output_file)])

    @classmethod
    def from_json(cls, json_file):
        """Create a list of datapoint from a json file."""
        with open(json_file) as f:
            return cls([DatapointFactory.from_json_line(line) for line in f if line.strip() not in ['{', '}']])


    def to_base_pairs_npy(self, path):
        """Writes the structure npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: structure matrix of pairs of 0-based indices.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])])
            >>> datapoints.to_base_pairs_npy('temp/base_pairs.npy')
            array([[[1, 2],
                    [3, 4]]], dtype=object)
        """

        arr = np.array([np.array(datapoint.paired_bases) for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_dms_npy(self, path):
        """Writes the dms npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: dms matrix of floats arrays.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])])
            >>> datapoints.to_dms_npy('temp/dms.npy')
            array([[1.0, 2.0, 3.0]], dtype=object)
        """

        arr = np.array([datapoint.dms for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_sequence_npy(self, path):
        """Writes the sequence npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: int-encoded sequence matrix

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0]),\
                                               Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])])
            >>> datapoints.to_sequence_npy('temp/sequence.npy')
            array([[1, 1, 2, 2, 3, 3],
                   [1, 1, 2, 2, 3, 3]], dtype=object)
        """

        arr = np.array([datapoint.embed_sequence() for datapoint in self.datapoints], dtype=object)
        np.save(path, arr)
        return arr


    def to_reference_npy(self, path):
        """Writes the reference npy file from the list of datapoints.

        Args:
            path (str): path to the npy file

        Returns:
            np.array: reference matrix of int-encoded.

        Examples:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])])
            >>> datapoints.to_reference_npy('temp/reference.npy')
            array(['reference'], dtype='<U9')
        """

        arr = np.array([datapoint.reference for datapoint in self.datapoints])
        np.save(path, arr)
        return arr


    def to_json(self, path) -> None:
        """Write a list of datapoints to a json file."""
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write('{\n')
            for idx, datapoint in enumerate(self.datapoints): # write the datapoints one by one to avoid memory issues
                f.write(str(datapoint)+ ',\n' if idx != len(self.datapoints)-1 else str(datapoint)+'\n')
            f.write('}\n')

    def to_pandas(self) -> pd.DataFrame:
        """Converts the list of datapoints into a pandas dataframe.

        Example:
            >>> datapoints = ListofDatapoints([Datapoint(reference='reference', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[1.0, 2.0, 3.0])])
            >>> datapoints.to_pandas()
               reference sequence      paired_bases              dms
            0  reference   AACCGG  [[1, 2], [3, 4]]  [1.0, 2.0, 3.0]
        """
        return pd.DataFrame([datapoint.to_flat_dict() for datapoint in self.datapoints])

    def filter_duplicates(self):
        """Filters out duplicate sequences.
        Only keep the first occurence of a sequence if all the other structures are the same.

        Examples:
            >>> datapoints = ListofDatapoints([ Datapoint(reference='ref1', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref2', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref3', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0,0,0,0,0,0]),\
                                Datapoint(reference='ref4', sequence='AUGGC', paired_bases=[[1, 2]], dms=[0,0,0,0,0]),\
                                Datapoint(reference='ref5', sequence='GCCUA', paired_bases=[[0, 4], [2, 3]], dms=[0,0,0,0,0]),\
                                Datapoint(reference='ref6', sequence='GCCUA', paired_bases=[[0, 4]], dms=[0,0,0,0,0]) ])
            >>> datapoints.filter_duplicates()
            >>> datapoints.datapoints
            [Datapoint('ref1', sequence='AACCGG', paired_bases=[[1, 2], [3, 4]], dms=[0, 0, 0, 0, 0, 0]), Datapoint('ref4', sequence='AUGGC', paired_bases=[[1, 2]], dms=[0, 0, 0, 0, 0])]
        """

        # Group datapoints by sequence
        unique_seqs = {}
        for i, datapoint in enumerate(self.datapoints):

            struct = np.array(datapoint.paired_bases)

            if datapoint.sequence not in unique_seqs.keys():
                unique_seqs[datapoint.sequence] = ([struct], [i])
            else:
                unique_seqs[datapoint.sequence][0].append(struct)
                unique_seqs[datapoint.sequence][1].append(i)

        # Only keep sequence duplicate that have the same pairing matrix
        idxs_to_remove = []
        for seq in unique_seqs.keys():
            if len(unique_seqs[seq]) > 1:
                matches = []
                for i in range(len(unique_seqs[seq][0])):
                    for j in range(i+1, len(unique_seqs[seq][0])):
                        matches.append(np.array_equal(unique_seqs[seq][0][i], unique_seqs[seq][0][j]))

                # Remove sequences from dataframe if f1 score is not 1.0, keep only one of the structures otherwise
                if not np.array(matches).all():
                    idxs_to_remove += unique_seqs[seq][1]
                else:
                    idxs_to_remove += unique_seqs[seq][1][1:]

        idxs_to_keep = list( set(range(len(self.datapoints)))-set(idxs_to_remove) )
        self.datapoints = [self.datapoints[i] for i in idxs_to_keep]



def base_pairs_into_matrix(base_pairs):
    """Turns a list of base pairs into a matrix of 1s and 0s.

    Args:
        base_pairs (list): list of base pairs, where each base pair is a tuple of 0-based two ints.

    Returns:
        matrix (np.array): matrix of 1s and 0s.

    Example:
        >>> from numpy import array, int64
        >>> base_pairs = [(0, 1), (2, 3), (4, 5)]
        >>> assert (base_pairs_into_matrix(base_pairs) ==  array([[0., 1., 0., 0., 0., 0.], \
                                                                [1., 0., 0., 0., 0., 0.], \
                                                                [0., 0., 0., 1., 0., 0.], \
                                                                [0., 0., 1., 0., 0., 0.], \
                                                                [0., 0., 0., 0., 0., 1.], \
                                                                [0., 0., 0., 0., 1., 0.]], dtype=int64)).all()

    """
    max_index = max([max(base_pair) for base_pair in base_pairs])
    arr = np.zeros((max_index + 1, max_index + 1))
    for base_pair in base_pairs:
        arr[base_pair[0], base_pair[1]] = 1
        arr[base_pair[1], base_pair[0]] = 1
    return arr


import datetime, os, json
from typing import Any
from .path import PathDatafolder

STRUCTURE_FROM_RNASTRUCTURE = 'RNAstructure'
STRUCTURE_FROM_SOURCE = 'source'
DMS_FROM_RNASTRUCTURE = 'RNAstructure'
DMS_FROM_SOURCE = 'source'

class InfoFileWriterTemplate(PathDatafolder):

    def __init__(self, name, root) -> None:
        super().__init__(name, root)

        self.info = {
            'name': self.name,
            'upload_date': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            'user': os.environ['USER'],
        }

    def write(self):
        """Write the info file and the readme file."""
        with open(self.get_info_file(), 'w') as f:
            json.dump(self.info, f, indent=4)

        with open(self.get_readme_file(), 'w') as f:
            f.write(f"""
---
license: mit
language:
  - en
tags:
  - chemistry
  - biology
author: Silvi Rouskin
pretty_name: {self.name}
---

""")

            for k, v in self.info.items():
                if k not in ['structure', 'about structure', 'DMS', 'about DMS']:
                    f.write(f"{k}: {v}\n\n")
            f.write(f"""

structure: {self.info['structure']} ({self.info['about structure']})

DMS: {self.info['DMS']} ({self.info['about DMS']})

"""



)



class InfoFileWriterFromDREEMoutput(InfoFileWriterTemplate):

    def __init__(self, name, root, path_in) -> None:
        super().__init__(name, root)

        file = json.load(open(path_in, 'r'))
        if 'user' in file:
            file['experimenter'] = file.pop('user')
        for k, v in file.items():
            if type(v) != dict:
                self.info[k] = v

        self.info['structure'] = STRUCTURE_FROM_RNASTRUCTURE
        self.info['about structure'] = 'Predicted using RNAstructure and the DMS reactivity data.'
        self.info['DMS'] = DMS_FROM_SOURCE
        self.info['about DMS'] = 'Measured experimentally using the attached experimental conditions.'


class InfoFileWriterFromCT(InfoFileWriterTemplate):

    def __init__(self, name, root) -> None:
        super().__init__(name, root)

        self.info['structure'] = STRUCTURE_FROM_SOURCE
        self.info['about structure'] = 'Read from source.'
        self.info['DMS'] = DMS_FROM_RNASTRUCTURE
        self.info['about DMS'] = 'Predicted using RNAstructure at 310K.'


class InfoFileWriterFromFasta(InfoFileWriterTemplate):

    def __init__(self, name, root) -> None:
        super().__init__(name, root)

        self.info['structure'] = STRUCTURE_FROM_RNASTRUCTURE
        self.info['about structure'] = 'Predicted using RNAstructure at 310K.'
        self.info['DMS'] = DMS_FROM_RNASTRUCTURE
        self.info['about DMS'] = 'Predicted using RNAstructure at 310K.'


def infoFileWriter(source, datafolder):
    if source == 'dreem_output':
        return InfoFileWriterFromDREEMoutput(datafolder.name, datafolder.path_out, datafolder.path_in)
    if source == 'ct':
        return InfoFileWriterFromCT(datafolder.name, datafolder.path_out)
    if source == 'fasta':
        return InfoFileWriterFromFasta(datafolder.name, datafolder.path_out)
    raise ValueError(f"Unknown source: {source}")

